/*
 * $Id: bgp_msg.c,v 1.2 1998/11/23 22:45:20 masaki Exp $
 */

#include <mrt.h>
#include <aspath.h>
#include <bgp.h>
#include <ctype.h>

extern char *s_origins[];

/* 
 * Process an MRT BGP update message. Fill in pp_attr
 * and ll_ann_prefixes and ll_with_prefixes.
 * XXX This function is used by programs other than mrtd
 */
int
bgp_process_update_msg (enum MRT_MSG_TYPES type, u_char * value, int length,
			gateway_t ** gateway_to, bgp_attr_t ** pp_attr,
			LINKED_LIST ** ll_with_prefixes,
			LINKED_LIST ** ll_ann_prefixes)
{
#ifdef HAVE_IPV6
    struct in6_addr addr;
#else
    u_long addr;
#endif /* HAVE_IPV6 */
    int as, ret;
    prefix_t *prefix;
    gateway_t *gateway_from = NULL;
    int plen = 4;
    int family = AF_INET;
    u_char *cp = value;
    int bgp4plus = DEFAULT_BGP4PLUS_VERSION;
    bgp_peer_t *peer;
    char str[MAXLINE];

#ifdef HAVE_IPV6
    if (type == MSG_PROTOCOL_BGP4PLUS) {
	plen = 16;
	family = AF_INET6;
    }
    if (type == MSG_PROTOCOL_BGP4PLUS_01) {
	plen = 16;
	family = AF_INET6;
	bgp4plus = 1;
    }
#endif /* HAVE_IPV6 */

    /* create a pseudo peer */
    peer = New (bgp_peer_t);
    peer->trace = trace_copy (MRT->trace);
    if (bgp4plus) {
        BIT_SET (peer->options, BGP_BGP4PLUS_01);
    }

    /* from */
    BGP_GET_SHORT (as, cp);
    BGP_GET_DATA (&addr, plen, cp);
    prefix = New_Prefix (family, &addr, plen * 8);
    if (is_any_addr (prefix)) {
#ifdef HAVE_IPV6
        if (prefix->family == AF_INET6)
           strcpy (str, "BGP4+");
        else
#endif /* HAVE_IPV6 */
        strcpy (str, "BGP");
    }
    else {
        gateway_from = add_gateway (prefix, as, NULL);

#ifdef HAVE_IPV6
        if (prefix->family == AF_INET6)
            sprintf (str, "BGP4+ %s", prefix_toa (prefix));
        else
#endif /* HAVE_IPV6 */
            sprintf (str, "BGP %s", prefix_toa (prefix));
    }
    set_trace (peer->trace, TRACE_PREPEND_STRING, str, 0);
    Deref_Prefix (prefix);

    /* to */
    BGP_GET_SHORT (as, cp);
    BGP_GET_DATA (&addr, plen, cp);
    prefix = New_Prefix (family, &addr, plen * 8);
    if (is_any_addr (prefix))
        *gateway_to = NULL;
    else
        *gateway_to = add_gateway (prefix, as, NULL);
    Deref_Prefix (prefix);

    length -= (cp - value);

    if ((ret = bgp_process_update_packet (cp, length, peer)) < 0) {
	/* code and subcode available */
    }
    else {
        /* in case of withdraw only */
        if (peer->attr == NULL)
	    peer->attr = bgp_new_attr (PROTO_BGP);
        peer->attr->gateway = gateway_from;
    }

    /* setting return values */
    *pp_attr = peer->attr;
    *ll_with_prefixes = peer->ll_withdraw;
    *ll_ann_prefixes = peer->ll_announce;

    Destroy_Trace (peer->trace);
    Delete (peer);

    return (ret);
}


/* 
 * Create MRT update message. Take in pointer to
 * buffer of allocated memory, and attr and with and
 * ann prefixes. Return pointer to end of buffer
 * filled in with MRT update message. Set size to size
 * filled in.
 */
u_char *
bgp_create_update_msg (enum MRT_MSG_TYPES type, int * size,
		       bgp_attr_t * attr,
		       LINKED_LIST * ll_ann_prefixes,
		       LINKED_LIST * ll_with_prefixes,
		       gateway_t *gateway_to)
{
    u_char *cp, *start_pdu, *end_pdu;
    int len;
    int plen = 4;
    int bgp4plus = DEFAULT_BGP4PLUS_VERSION;

#ifdef HAVE_IPV6
    if (type == MSG_PROTOCOL_BGP4PLUS) {
	assert (attr->gateway == NULL ||
		attr->gateway->prefix->family == AF_INET6);
	plen = 16;
    }
    if (type == MSG_PROTOCOL_BGP4PLUS_01) {
	assert (attr->gateway == NULL ||
		attr->gateway->prefix->family == AF_INET6);
	plen = 16;
	bgp4plus = 1;
    }
#endif /* HAVE_IPV6 */

    cp = start_pdu = NewArray (u_char, MAX_MSG_SIZE);
    end_pdu = start_pdu + MAX_MSG_SIZE;

    /* from */
    if (attr == NULL || attr->gateway == NULL) {
	memset (cp, 0, plen + 2);
	cp += (plen + 2);
    }
    else {
	BGP_PUT_SHORT (attr->gateway->AS, cp);
        BGP_PUT_DATA (prefix_tochar (attr->gateway->prefix), plen, cp);
    }

    /* to */
    if (gateway_to == NULL) {
	memset (cp, 0, plen + 2);
	cp += (plen + 2);
    }
    else {
	BGP_PUT_SHORT (gateway_to->AS, cp);
        BGP_PUT_DATA (prefix_tochar (gateway_to->prefix), plen, cp);
    }

    len = bgp_create_pdu (ll_with_prefixes, ll_ann_prefixes, attr,
			  NULL, NULL, cp, end_pdu - cp, bgp4plus);

    if (len > 0) {
	*size = (cp + len) - start_pdu;
    }

    return (start_pdu);
}


void 
bgp_process_sync_msg (enum MRT_MSG_TYPES type, u_char *msg, int length, 
		      int *viewno, char *filename)
{
    *viewno = 0;
    BGP_GET_SHORT (*viewno, msg);
    length -= 2;
    strncpy (filename, (char *)msg, length);
    filename[length] = '\0';
}


void 
bgp_process_state_msg (enum MRT_MSG_TYPES type, u_char *msg, int length, 
		       gateway_t ** gateway,
		       u_short * old_state, u_short * new_state)
{
    u_char *cp, ip[16];
    u_short as;
    prefix_t *prefix;
    int plen = 4;
    int family = AF_INET;
    gateway_t *gateway_from;
#ifdef notdef
    gateway_t *gateway_to;
#endif /* notdef */

#ifdef HAVE_IPV6
    if (type == MSG_PROTOCOL_BGP4PLUS || type == MSG_PROTOCOL_BGP4PLUS_01) {
	plen = 16;
        family = AF_INET6;
    }
#endif /* HAVE_IPV6 */

    cp = msg;

    /* from */
    BGP_GET_SHORT (as, cp);
    BGP_GET_DATA (ip, plen, cp);
    prefix = New_Prefix (family, ip, plen * 8);
    gateway_from = add_gateway (prefix, as, NULL);
    Deref_Prefix (prefix);

#ifdef notdef
    /* to */
    BGP_GET_SHORT (as, cp);
    BGP_GET_DATA (ip, plen, cp);
    prefix = New_Prefix (family, ip, plen * 8);
    gateway_to = add_gateway (prefix, as, NULL);
    Deref_Prefix (prefix);
#endif

    BGP_GET_SHORT (*old_state, cp);
    BGP_GET_SHORT (*new_state, cp);

    *gateway = gateway_from;
    return;
}


char *
bgp_attr_toa (bgp_attr_t * attr)
{
    char *tmpx;
    char *cp;

    THREAD_SPECIFIC_STRAGE (tmpx);

    sprintf (tmpx, "%s|%s|%s|%d|%d|%s|%s|",
	     aspath_toa (attr->aspath),
	     origin2string (attr->origin),
	     attr->nexthop ? prefix_toa (attr->nexthop->prefix) : "none",
	     (int) attr->local_pref,
	     (int) attr->multiexit,
	     community_toa (attr->community),
	     BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG) ? "AG" : "NAG");

    /* I don't think we can call prefix_toa twice in the same call */
    cp = tmpx + strlen (tmpx);
    sprintf (cp, "%s|",  attr->aggregator.prefix ? 
			     prefix_toa (attr->aggregator.prefix) : "");

    return (tmpx);
}


void
bgp_print_attr (bgp_attr_t * attr)
{
    assert (attr);

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN))
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_ORIGIN),
			    origin2string (attr->origin));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH))
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_ASPATH),
			    aspath_toa (attr->aspath));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP))
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->nexthop->prefix));
#ifdef HAVE_IPV6
    if (attr->link_local)
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->link_local->prefix));
    if (attr->nexthop4)
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->nexthop4->prefix));
#endif /* HAVE_IPV6 */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC))
	printf ("%s: %ld\n", bgptype2string (PA4_TYPE_METRIC),
			     attr->multiexit);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF))
	printf ("%s: %ld\n", bgptype2string (PA4_TYPE_LOCALPREF),
			     attr->local_pref);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG))
	printf ("%s\n", bgptype2string (PA4_TYPE_ATOMICAGG));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR))
	printf ("%s: AS%d %s\n", bgptype2string (PA4_TYPE_AGGREGATOR),
				 attr->aggregator.as,
				 prefix_toa (attr->aggregator.prefix));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGINATOR_ID))
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_ORIGINATOR_ID),
			    prefix_toa (attr->originator));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_CLUSTER_LIST))
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_CLUSTER_LIST),
			    cluster_list_toa (attr->cluster_list));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY))
	printf ("%s: %s\n", bgptype2string (PA4_TYPE_COMMUNITY),
			    community_toa (attr->community));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA))
	printf ("%s: AS%d %ld\n", bgptype2string (PA4_TYPE_DPA),	
				  attr->dpa.as, attr->dpa.value);
}


/* just i'm lazy to write a code to check buffer overflow */
void
bgp_uii_attr (uii_connection_t *uii, bgp_attr_t * attr)
{
    assert (attr);

    if (attr->gateway)
        uii_add_bulk_output (uii, "FROM: %s AS%d\n",
                                 prefix_toa (attr->gateway->prefix),
                                 attr->gateway->AS);

    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGIN))
	uii_add_bulk_output (uii, "%s: %s\n", bgptype2string (PA4_TYPE_ORIGIN),
			    origin2string (attr->origin));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ASPATH))
	uii_add_bulk_output (uii, "%s: %s\n", bgptype2string (PA4_TYPE_ASPATH),
			    aspath_toa (attr->aspath));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_NEXTHOP)) {
	if (attr->nexthop2)
	    uii_add_bulk_output (uii, "%s: %s via %s\n", 
			    bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->nexthop->prefix),
			    prefix_toa (attr->nexthop2->prefix));
	else
	    uii_add_bulk_output (uii, "%s: %s\n", 
			    bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->nexthop->prefix));
    }
#ifdef HAVE_IPV6
    if (attr->link_local)
	uii_add_bulk_output (uii, "%s: %s\n", bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->link_local->prefix));
    if (attr->nexthop4)
	uii_add_bulk_output (uii, "%s: %s\n", bgptype2string (PA4_TYPE_NEXTHOP),
			    prefix_toa (attr->nexthop4->prefix));
#endif /* HAVE_IPV6 */
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_METRIC))
	uii_add_bulk_output (uii, "%s: %ld\n", bgptype2string (PA4_TYPE_METRIC),
			     attr->multiexit);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_LOCALPREF))
	uii_add_bulk_output (uii, "%s: %ld\n", 
			     bgptype2string (PA4_TYPE_LOCALPREF),
			     attr->local_pref);
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ATOMICAGG))
	uii_add_bulk_output (uii, "%s\n", bgptype2string (PA4_TYPE_ATOMICAGG));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_AGGREGATOR))
	uii_add_bulk_output (uii, "%s: AS%d %s\n", 
				 bgptype2string (PA4_TYPE_AGGREGATOR),
				 attr->aggregator.as,
				 prefix_toa (attr->aggregator.prefix));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_ORIGINATOR_ID))
	uii_add_bulk_output (uii, "%s: %s\n", 
			    bgptype2string (PA4_TYPE_ORIGINATOR_ID),
			    prefix_toa (attr->originator));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_CLUSTER_LIST))
	uii_add_bulk_output (uii, "%s: %s\n", 
			    bgptype2string (PA4_TYPE_CLUSTER_LIST),
			    cluster_list_toa (attr->cluster_list));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_COMMUNITY))
	uii_add_bulk_output (uii, "%s: %s\n", 
			    bgptype2string (PA4_TYPE_COMMUNITY),
			    community_toa (attr->community));
    if (BGP4_BIT_TEST (attr->attribs, PA4_TYPE_DPA))
	uii_add_bulk_output (uii, "%s: AS%d %ld\n", 
			          bgptype2string (PA4_TYPE_DPA),	
				  attr->dpa.as, attr->dpa.value);
}


int
bgp_scan_attr (char *line, bgp_attr_t *attr, trace_t *tr)
{
    char *cp = line;
    int code;

    assert (line);

    if ((code = string2bgptype (&cp)) <= 0)
	return (0);

    if (code == PA4_TYPE_ATOMICAGG) {
	/* check to make sure */
	if (*cp == '\0' || isspace (*cp)) {
    	    BGP4_BIT_SET (attr->attribs, code);
	    bgp_trace_attr (TR_PACKET, tr, attr, code);
	    return (code);
	}
	return (0);
    }

    if (*cp != ':' && !isspace (*cp)) {
	return (0);
    }

    do {
        cp++;
    } while (isspace (*cp));

    switch (code) {

    case PA4_TYPE_ORIGIN: {
	int i;
	for (i = 0; i < 3; i++) {
	    if (strncasecmp (cp, s_origins[i], strlen (s_origins[i])) == 0) {
		attr->origin = i;
    	        BGP4_BIT_SET (attr->attribs, code);
	        bgp_trace_attr (TR_PACKET, tr, attr, code);
		return (code);
	    }
	}
	return (-1);
    }
    case PA4_TYPE_ASPATH: {
	aspath_t *aspath;
	aspath = aspth_from_string (cp);
	if (aspath == NULL) {
	    return (-1);
	}
	if (attr->aspath)
	    Delete_ASPATH (attr->aspath);
	attr->aspath = aspath;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
        return (code);
    }
    case PA4_TYPE_NEXTHOP: {
	prefix_t *prefix;
#ifdef HAVE_IPV6
	prefix = ascii2prefix (0, cp);
	if (prefix == NULL) {
	    return (-1);
	}
	if (prefix->family == AF_INET6) {
	    if (attr->nexthop && attr->nexthop->prefix->family == AF_INET) {
		attr->nexthop4 = attr->nexthop;  /* save */
                attr->nexthop = NULL;
	    }
	    if (IN6_IS_ADDR_LINKLOCAL (prefix_toaddr6 (prefix))) {
		attr->link_local = add_nexthop (prefix, NULL);
		if (attr->nexthop == NULL)
		    attr->nexthop = ref_nexthop (attr->link_local);
	    }
	    else {
		if (attr->nexthop)
		    deref_nexthop (attr->nexthop);
		attr->nexthop = add_nexthop (prefix, NULL);
	    }
	}
	else {
	    if (attr->nexthop && attr->nexthop->prefix->family == AF_INET6) {
		if (attr->nexthop4)
		    deref_nexthop (attr->nexthop4);
		attr->nexthop4 = add_nexthop (prefix, NULL);
	    }
	    else {
		if (attr->nexthop)
		    deref_nexthop (attr->nexthop);
		attr->nexthop = add_nexthop (prefix, NULL);
	    }
	}
    	BGP4_BIT_SET (attr->attribs, code);
	trace (TR_PACKET, tr, "  %s: %s\n", bgptype2string (PA4_TYPE_NEXTHOP),
				       prefix_toa (prefix));
#else
	prefix = ascii2prefix (AF_INET, cp);
	if (prefix == NULL) {
	    return (-1);
	}
	if (attr->nexthop)
	    deref_nexthop (attr->nexthop);
	attr->nexthop = add_nexthop (prefix, NULL);
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
#endif /* HAVE_IPV6 */
        return (code);
    }
    case  PA4_TYPE_METRIC: {
	int value;
	if (sscanf (cp, "%d", &value) < 1)
	    return (-1);
	attr->multiexit = value;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    case PA4_TYPE_LOCALPREF: {
	int value;
	if (sscanf (cp, "%d", &value) < 1)
	    return (-1);
	attr->local_pref = value;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    case PA4_TYPE_AGGREGATOR: {
	int as; char str[MAXLINE];
	if (sscanf (cp, "AS%d %s", &as, str) < 2)
	    return (-1);
	attr->aggregator.as = as;
	if (attr->aggregator.prefix)
	    Deref_Prefix (attr->aggregator.prefix);
	attr->aggregator.prefix = ascii2prefix (AF_INET, str);
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    case PA4_TYPE_COMMUNITY: {
	community_t *community;
	community = community_from_string (cp);
	if (community == NULL)
	    return (-1);
	if (attr->community)
	    Delete_community (attr->community);
	attr->community = community;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    case PA4_TYPE_ORIGINATOR_ID: {
	prefix_t *prefix;
	prefix = ascii2prefix (AF_INET, cp);
	if (prefix == NULL)
	    return (-1);
	if (attr->originator)
	    Deref_Prefix (attr->originator);
	attr->originator = prefix;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    case PA4_TYPE_CLUSTER_LIST: {
	cluster_list_t *cluster_list;
	cluster_list = cluster_list_from_string (cp);
	if (cluster_list == NULL)
	    return (-1);
	if (attr->cluster_list)
	    Delete_cluster_list (attr->cluster_list);
	attr->cluster_list = cluster_list;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    case PA4_TYPE_DPA: {
	int as, value;
	if (sscanf (cp, "AS%d %d", &as, &value) < 2)
	    return (-1);
	attr->dpa.as = as;
	attr->dpa.value = value;
    	BGP4_BIT_SET (attr->attribs, code);
	bgp_trace_attr (TR_PACKET, tr, attr, code);
	return (code);
    }
    default:
	break;
    }
    return (-1);
}


