/* 
 * $Id: bgpsim.c,v 1.19 1998/11/23 22:50:00 masaki Exp $
 */

#include <ctype.h>
#include <mrt.h>
#include <config_file.h>
#include <protoconf.h>
#include "bgpsim.h"


simulation_t *SIMULATION;
io_t *IO;
gateway_t *local_gateway;
trace_t *default_trace;
#ifdef HAVE_IPV6
gateway_t *local_gateway6;
#endif /* HAVE_IPV6 */

int show_simulation (uii_connection_t * uii);
void start_simulation (trace_t * trace);
int stop_simulation_schedule (uii_connection_t * uii);

typedef struct _config_bgpsim_t {
    struct _network_t *network;
} config_bgpsim_t;

static config_bgpsim_t *CONFIG_BGPSIM;


static network_t *
config_find_network (int num)
{
    network_t *network;

    LL_Iterate (SIMULATION->ll_networks, network) {
	if (network->num == num)
	    return (network);
    }

    return (NULL);
}


static void
get_config_network_list (int num)
{
    network_t *network;
    range_t *range;
    int i;

	if ((network = config_find_network (num)) == NULL)
		return;

    config_add_output ("network-list %d\n", num);
    LL_Iterate (network->ll_range, range) {
	config_add_output ("  range %s %s\n", prefix_toax (range->start), 
					      prefix_toa (range->end));
    }
    config_add_output ("  stability %d\n", network->stability);
    config_add_output ("  change %d\n", network->change_interval);
    config_add_output ("  route-map");
    for (i = 1; i <= network->max_set; i++) {
	config_add_output (" %d", network->sets[i]);
    }
    config_add_output ("\n");
}


static int 
config_network_list (uii_connection_t * uii, int num)
{
    network_t *network;

	if ((network = config_find_network (num)) == NULL) {
            network = New (network_t);
	    network->num = num;
	    memset (network->sets, 0, sizeof (network->sets));
	    LL_Add (SIMULATION->ll_networks, network);
	    CONFIG_BGPSIM->network = network;
	    uii->state = UII_CONFIG_NETWORK_LIST;
	    config_add_module (0, "network-list", get_config_network_list,
			      (void *)num);
	    return (1);
        }
	else {
	    trace (ERROR, CONFIG.trace,
		   "*ERROR* CONFIG network-list %d already defined\n", num);
    	    return (-1);
	}
}


static int
network_list_range (uii_connection_t * uii, prefix_t *start, prefix_t *end)
{
	range_t *range;

	if (start->family != end->family) {
	    Deref_Prefix (start);
	    Deref_Prefix (end);
	    return (-1);
	}

	if (CONFIG_BGPSIM->network->ll_range == NULL)
            CONFIG_BGPSIM->network->ll_range = LL_Create (0);

	range = New (range_t);
	range->start = start;
	range->end = end;
	LL_Add (CONFIG_BGPSIM->network->ll_range, range);
	return (1);
}


static int
network_list_stability (uii_connection_t * uii, int stability)
{
    char tmpx[MAXLINE];

    /* stability */
        sprintf (tmpx, "Network Stability %d", CONFIG_BGPSIM->network->num);
	CONFIG_BGPSIM->network->stability = stability;
	CONFIG_BGPSIM->network->timer_stability =
	New_Timer (network_schedule_stability,
		   CONFIG_BGPSIM->network->stability, tmpx, 
		   CONFIG_BGPSIM->network);
	return (1);
}


static int 
network_list_load_file (uii_connection_t *uii, char *filename) {
  
  /* don't allow both range and filename to be configured */

  /*if (CONFIG_BGPSIM->network->ll_routes == NULL)
    CONFIG_BGPSIM->network->ll_routes = LL_Create (0);*/

  CONFIG_BGPSIM->network->filename = filename;

  return (1);
}


static int
network_list_change (uii_connection_t * uii, int change)
{
    char tmpx[MAXLINE];

	sprintf (tmpx, "Network Change %d", CONFIG_BGPSIM->network->num);
	CONFIG_BGPSIM->network->change_interval = change;
	CONFIG_BGPSIM->network->timer_change = 
		New_Timer (network_schedule_change,
			   CONFIG_BGPSIM->network->change_interval, tmpx,
			   CONFIG_BGPSIM->network);
	return (1);
}


static int
network_list_set (uii_connection_t * uii, char *arg)
{
    char *list = arg;
    char tmpx[MAXLINE], *cp;

	int set;

	while ((cp = uii_parse_line2 (&list, tmpx)) != NULL) {

	    if (!isdigit (cp[0])) {
		Delete (arg);
		return (0);
	    }

	    set = atoi (cp);

	    if (set <= 0) {
		trace (ERROR, CONFIG.trace, "CONFIG set number should be > 0\n");
		Delete (arg);
		return (-1);
	    }
	    if (CONFIG_BGPSIM->network->max_set >= MAX_SETS - 1) {
		trace (ERROR, CONFIG.trace, 
		       "CONFIG too many sets %d (should be < %d)\n", 
		       CONFIG_BGPSIM->network->max_set, MAX_SETS);
		Delete (arg);
		return (-1);
	    }
	    CONFIG_BGPSIM->network->sets[++CONFIG_BGPSIM->network->max_set] 
		= set;
	    trace (NORM, CONFIG.trace, 
		   "CONFIG new set %d for network-list %d\n", 
		   set, CONFIG_BGPSIM->network->num);
	}
	Delete (arg);
	return (1);
}


#ifdef notdef
static int
route_set_compare (u_int *a, u_int *b)
{
   return (*a - *b);
}
#endif


/* neighbor xx.xx.xx.xx stability %d */
static int 
config_router_neighbor_stability (uii_connection_t *uii,
				  prefix_t *prefix, int stability) {
  bgp_peer_t *peer;
  peer_flap_t *flap;

  if ((peer = Find_BGP_Peer_ByPrefix (prefix)) == NULL) {
    config_notice (TR_ERROR, uii, "No peer %s\n", prefix_toa (prefix));
    Deref_Prefix (prefix);
    return (-1);
  }
  Deref_Prefix (prefix);

  if (SIMULATION->ll_peer_flaps == NULL)
    SIMULATION->ll_peer_flaps =  LL_Create (0);

  LL_Iterate (SIMULATION->ll_peer_flaps, flap) {
    if (flap->peer == peer) {
      Timer_Set_Time (flap->timer, stability);
      return (1);
    }
  }

  flap = New (peer_flap_t);
  flap->peer = peer;
  flap->timer = New_Timer (bgpsim_flap_peer, stability, "Flap Peer", peer);

  LL_Add (SIMULATION->ll_peer_flaps, flap);
  return (1);
}


static void 
add_bgpsim_config () {

    CONFIG_BGPSIM = New (config_bgpsim_t);

    uii_add_command2 (UII_CONFIG, COMMAND_NORM, "network-list %d", 
		config_network_list, "Defines a network list");

    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "range %m %m", 
		     network_list_range, "Prefix range to generate");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "stability %d", 
		     network_list_stability, "Flapping interval (sec)");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "change %d", 
		     network_list_change, "Rotates route-maps");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "route-map %S",
		      network_list_set, "Lists route-map to use");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "file %sfilename",
		      network_list_load_file, 
		      "Load routes from routing table dump");
    uii_add_command2 (UII_CONFIG_NETWORK_LIST, COMMAND_NORM, "file %sfilename",
		      network_list_load_file, 
		      "Load routes from routing table dump");
    uii_add_command2 (UII_CONFIG_ROUTER_BGP, COMMAND_NORM,
		      "neighbor %m stability %d", 
		      config_router_neighbor_stability,
		      "Set stability of TCP peering session with neighbor");
}


void 
main (int argc, char *argv[])
{
    char c;
    extern char *optarg;	/* getopt stuff */
    extern int optind;		/* getopt stuff */
    int errors = 0;
    u_long id = htonl (DEFAULT_MY_ID);

    char *usage = "Usage: bgpsim [-i router_id] [-f file] [-v]\n";
    char *config_file = "./bgpsim.conf";
    default_trace = New_Trace ();

    /* must be root -- I guess the BGP code requires this somewhere?? */
    if (getuid ()) {
      fprintf (stderr, "bgpsim: must be root\n");
      exit (1);
    }

    /* init simulation */
    SIMULATION = New (simulation_t);
    SIMULATION->ll_networks = LL_Create (0);
    BGPSIM_TRANSPARENT = 1;

    while ((c = getopt (argc, argv, "vf:i:s")) != -1)
      switch (c) {
	case 'v':		/* verbose */
	    set_trace (default_trace, TRACE_FLAGS, TR_ALL,
		       TRACE_LOGFILE, "stdout",
		       NULL);
	    break;
	case 'f':		/* config file */
	    config_file = strdup (optarg);
	    break;
	case 'i':
	    if (inet_pton (AF_INET, optarg, &id) > 0)
	        break;
        case 's':  /* by default, bgpsim is transparent. set nexthop, origin... */
	  BGPSIM_TRANSPARENT = 0;
	  break;
	case 'h':
	default:
	    errors++;
	    break;
	}

    if (errors) {
      fprintf (stderr, usage);
	printf ("\nBGPSIM version (%s) compiled on %s\n\n",
		MRT_VERSION, __DATE__);
	exit (0);
    }

    init_mrt (default_trace);
    init_uii (default_trace);
    init_mrt_reboot (argc, argv);

    init_interfaces (default_trace); 
    init_mrtd_config (default_trace);
    config_bgp_init ();
    config_rtmap_init ();
    /* config_rip_init (); */
    add_bgpsim_config ();

    init_BGP (default_trace);
    set_BGP (BGP_MY_ID, id, 0);
    /* drop all updates */
    set_BGP (BGP_RECV_UPDATE_FN, NULL, 0);
    set_BGP (BGP_STATE_CHANGE_FN, NULL, 0);

    if (config_from_file (default_trace, config_file) < 0) {
      UII->initial_state = 1;   /* no password */
      config_create_default ();
    }

    local_gateway = add_gateway (ascii2prefix (AF_INET, "127.0.0.1"), 
				 BGP->my_as, NULL);
#ifdef HAVE_IPV6
    local_gateway6 = add_gateway (ascii2prefix (AF_INET6, "::1"), 
				  BGP->my_as, NULL);
#endif /* HAVE_IPV6 */

    set_uii (UII, UII_PROMPT, UII_NORMAL, "Bgpsim> ", 0);
    /* set_uii (UII, UII_PROMPT, UII_CONFIG, "Bgpsim-Config> ", 0); */

    /*   signal (SIGINT, sbgp_quit); */

    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show simulation", 
		     show_simulation, "Show status of simulation");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "stop", 
		     stop_simulation_schedule, "Stop simulation");

    listen_uii2 ("mrtd");

    start_simulation (default_trace);
    mrt_main_loop ();
    exit (0);
}
